DEFAULT_VEGA = {
    "$schema": "https://vega.github.io/schema/vega/v5.json",
    "width": 800,
    "height": 200,
    "padding": 5,
    "data": [{"name": "table", "values": []}],
    # "signals": [
    #     {
    #         "name": "tooltip",
    #         "value": {},
    #         "on": [
    #             {"events": "rect:mouseover", "update": "datum"},
    #             {"events": "rect:mouseout", "update": "{}"},
    #         ],
    #     }
    # ],
    "marks": [],
}
